/*
 * Bilateral Filter With Regression.
 *      LA FONCTION DE CHARLES
 *      la seule faite par lui
 *      dans tout ce paquet.
 */

#include "SparseFilter.h"
#include "WeightMap.h"
#include "Utils.h"
#include <algorithm>
#include <cmath>
#include <cassert>

using std::max;

namespace sparse_filter {
    
    namespace {
        
        inline int SymmetricCoordinate(int pos, int size) {
            if (pos < 0) pos = -pos - 1;
            if (pos >= 2 * size) pos %= 2 * size;
            if (pos >= size) pos = 2 * size - 1 - pos;
            return pos;
        }
        
        void PadSymmetric(const Image &src, int pad_before, int pad_after, Image *dst) {
            int pad_total = pad_before + pad_after;
            dst->Init(src.rows() + pad_total, src.columns() + pad_total, src.channels());
            for (int row = 0; row < dst->rows(); ++row) {
                for (int col = 0; col < dst->columns(); ++col) {
                    for (int chan = 0; chan < dst->channels(); ++chan) {
                        dst->val(col, row, chan) = src.val(SymmetricCoordinate(col - pad_before, src.columns()),
                                                           SymmetricCoordinate(row - pad_before, src.rows()),
                                                           chan);
                    }
                }
            }
        }
        
        void ExtractPatch(const Image &src, int pr, int pc, Image *dst) {
            // src is padded, so (pr, pc) becomes the upper left pixel
            for (int row = 0; row < dst->rows(); ++row) {
                for (int col = 0; col < dst->columns(); ++col) {
                    for (int chan = 0; chan < dst->channels(); ++chan) {
                        dst->val(col, row, chan) = src.val(pc + col, pr + row, chan);
                    }
                }
            }
        }
        
        
        
        
        void BilateralWeight(const Image &g, Image *k, int r, float sigma_d2, float sigma_r2) {
            
            for (int row = 0; row < g.rows(); ++row) {
                for (int col = 0; col < g.columns(); ++col) {
                    
                    float x = 0.f;
                    
                    for (int chan = 0; chan < g.channels(); ++chan) {
                        float y = g.val(col, row, chan) - g.val(r, r, chan);
                        x += y * y;
                    }
                    
                    x /= 2 * sigma_r2;
                    x += ((row - r) * (row - r) + (col - r) * (col - r)) / (2 * sigma_d2);
                    k->val(col, row) = utils::fastexp(-x);
                }
            }
        }
        
        
        
    }  // namespace
    
    
    
    void BilateralFilter(const Image& source, float sigma_d2, float sigma_r2, int r, Image* output) {
        
        // ------------------------------------------------------------------------------ //
        // Pseudo-code du filtre bilatéral avec régression :
        /*
         1. En dehors de la boucle :
         ---------------------------
         
         1.a) Construction du noyau spatial
         
         Gaussienne spatialle sur un rayon r*r et de variance sigma_d2
         
         1.b) Gestion des bords
         
         Recopie en miroir de l'image d'entrée sur un nombre r de pixels
         de chaque côté de l'image
         
         1.c) Allocation de mémoire pour les variables...
         
         
         2. Dans la boucle : POUR CHAQUE PIXEL (h,w)
         -------------------------------------------
         
         2.a) Construction du noyan gaussien en intensité
         
         Voisinage dans r*r moins la valeur en (h,w) (valeur centrale)
         auquel on applique la fonction gaussienne de paramètre sigma_r2 (variance)
         
         2.b) Calcul des poids spatiaux du bilatéral
         
         Multiplication point à point des poids spatiaux et d'intensité.
         
         2.c) Calcul des valeurs "SIGMA" nécessaires :
         
         Voir mon document sur l'accélération du filtre bilatéral avec régression,
         on utilise ici la solution analytique et on ne calcule que le paramètre c
         du plan de régression (seul nécessaire), avec la méthode de Cramer.
         
         2.d) Résolution du système
         
         Comme je le dis juste au-dessus.
         
         2.e) L'image de sortie prend le paramètre c à la position (h,w)
         
         FIN DE L'ALGO
         */
        // ------------------------------------------------------------------------------ //
        
        
        
        // Initialization of the output image
        output->Init(source.rows(), source.columns(), source.channels());
        output->Reset();
        
        // Some variables
        int pr, pc;                         // coordinates of the central pixel
        int s = 2 * r + 1;                  // patch size
        Image y(s, s, source.channels());   // extracted patch
        Image k(s, s);                      // weights
        Image source_pad;                   // source image with padding
        
        // Some other variables
        float total_weight = 0.f;
        float yt[];
        int row, col, chan;
        
        // Image padding
        PadSymmetric(source, r, r, &source_pad); // +++ Utiliser source_pad dans la suite
        
        
        
        // Loop over the whole image
        for(pr=0;pr<source.rows();pr++){
            for(pc=0;pc<source.columns();pc++){
                
                // Patch centered in (pr,pc)
                ExtractPatch(source_pad, pr, pc, &y);
                
                // Computation of the bilateral weights
                BilateralWeight(y, &k, r, sigma_d2, sigma_r2);
                
                // Compute the filtered value at (pr,pc)
                for (chan = 0; chan < y.channels(); ++chan) {
                    yt[chan] = 0.f;
                }
                for (row = 0; row < y.rows(); ++row) {
                    for (col = 0; col < y.columns(); ++col) {
                        total_weight += k.val(col, row);
                        for (chan = 0; chan < y.channels(); ++chan) {
                            yt[chan] += y.val(col, row, chan) * k.val(col, row);
                        }
                    }
                }
                for (chan = 0; chan < y.channels(); ++chan) {
                    yt[chan] /= total_weight;
                }
                
                
                
                //ComputeAverage(y, k, yt);
                /*
                 void ComputeAverage(const Image &y, const Image &k, float yt[]) {
                 float total_weight = 0.f;
                 for (int chan = 0; chan < y.channels(); ++chan) {
                 yt[chan] = 0.f;
                 }
                 for (int row = 0; row < y.rows(); ++row) {
                 for (int col = 0; col < y.columns(); ++col) {
                 total_weight += k.val(col, row);
                 for (int chan = 0; chan < y.channels(); ++chan) {
                 yt[chan] += y.val(col, row, chan) * k.val(col, row);
                 }
                 }
                 }
                 for (int chan = 0; chan < y.channels(); ++chan) {
                 yt[chan] /= total_weight;
                 }
                 }
                 */
                
                
                
            }
        }
        
        
        
        
        
        
        
        /*
         // initializations and internal variables
         
         output->Init(source.rows(), source.columns(), source.channels());
         output->Reset();
         WeightMap agg_weights(source.rows(), source.columns());
         
         
         int pr, pc;  // coordinates of the central pixel
         int s = 2 * r + 1;
         Image y(s, s, source.channels());  // extracted patch
         float yt[source.channels()];  // average value
         Image k(s, s);  // weight function
         Image source_pad;
         
         
         // image padding
         PadSymmetric(source, r, r, &source_pad);
         
         
         // main loop
         while (agg_weights.Minimum() < threshold) {
         
         agg_weights.FindMinimum(&pr, &pc);
         ExtractPatch(source_pad, pr, pc, &y);
         BilateralWeight(y, &k, r, sigma_d2, sigma_r2);
         ComputeAverage(y, k, yt);
         AggregateValue(yt, k, pr - r, pc - r, output);
         //    SquarePatch(&k);
         agg_weights.IncreaseWeights(k, pr - r, pc - r);
         
         }
         
         Normalize(agg_weights, output);
         */
    }
    
    
    
}  // namespace da3d





